require "disiid_user/version"
require 'active_resource'

## Usage:
#
# in config/initializers/disid.rb:
#
#   DisiidUser::RemoteUser.site = 'http://id.provider.com'
#   DisiidUser::RemoteUser.auth_token = 'a secret token goes here'
#
#   # if resource name is different from 'user' (default)
#   DisiidUser::RemoteUser.element_name = 'user'
#
#
# in app/models/user.rb:
#
#   class User < ActiveRecord::Base
#     include DisiidUser
#     ...
#   end
#
module DisiidUser
  
  def self.included(klass)
    klass.class_eval do
      extend ClassMethods
      scope :with_local_role, lambda { |role| where('role & ? > 0', 2**LOCAL_ROLES.index(role.to_s)) }
    end
  end
  
  ########################################################################
  # Class Methods
  #
  module ClassMethods
    # Returns a list of users as ActiveResource objects that match given query params, e.g.
    #   User.search(:q => "dude")            # => [{name: 'dude', ...}, {...}]
    #   User.search(:position => 'whatever)  # => []
    def search(args = {})
      args.merge! :auth_token => DisiidUser::RemoteUser.auth_token
      DisiidUser::RemoteUser.find :all, :params => args
    end
    
    # Same as #search but returns just an integer, a number of how many search result items
    # the #search method would return (give the same arguments)
    #   User.remote_count(:q => "dude")                    # => 1
    #   User.remote_count(:position => 'technical_staff')  # => 13
    def remote_count(args = {})
      args.merge! :auth_token => DisiidUser::RemoteUser.auth_token
      DisiidUser::RemoteUser.get :count, args
    end
    
    # Creates a new user object with the identity attribute assigned
    # and returns a not presisted record. 
    # You'll need to call .save or .save! on the returned result 
    # if you want to persist it in a local database.
    #
    # Note: your local :users table should :identity_url string attribute for this to work.
    def build_from_identity_url(url)
      user = new
      user.identity_url = url
      user
    end
    
    alias :build_from_identity :build_from_identity_url
    
    # Looks for a local user record with a given UUID by matching its local attribute :identity_url.
    # Creates a new user if it doesn't exist locally but is present remotely using #build_from_identity method.
    # This method will throw ActiveRecord::RecordNotFound exception if persistance process didn't go well.
    def find_local_or_remote_by_uuid!(id)
      user = where('identity_url LIKE ?', "%#{id}").first
      if !user && DisiidUser::RemoteUser.exists?(id, params: {auth_token: DisiidUser::RemoteUser.auth_token})
        # DisiidUser::RemoteUser.site is a URI::HTTP object, not a string
        user_url = DisiidUser::RemoteUser.site.merge("#{DisiidUser::RemoteUser.collection_name}/#{id}").to_s
        user = build_from_identity(user_url)
        user.save!
      end
      user
    rescue
      raise ::ActiveRecord::RecordNotFound
    end
  end
  
  def self.version_string
    "DisiidUser version #{VERSION}"
  end
  
  ########################################################################
  # Check roles of user
  #
  
  # list of local roles
  LOCAL_ROLES = %w(admin manager) unless defined? LOCAL_ROLES
  
  # list of methid name that could be called from User instance to get data
  # from the RemoteUser instance
  REMOTE_ATTRIBUTES = %w(email first_name last_name homepage
                         nickname homepage position office_address office_num 
                         office_phone office_fax
                         provider ada_id okkam_id unitn_uid
                         photo photo_normal photo_thumb hide_photo
                         remote_created_at remote_updated_at) unless defined? REMOTE_ATTRIBUTES
  
  # List of roles for local setup
  def local_roles
    LOCAL_ROLES.reject { |r| ((role || 0) & 2**LOCAL_ROLES.index(r)).zero? }
  end
  
  # Add roles to the local setup
  # all of the following will work: 
  #   user.local_roles = :admin
  #   user.local_roles = [:admin, :manager]
  #   user.local_roles = 'admin'
  #   user.local_roles = 'admin', 'manager' # w/o brackets
  def local_roles=(*roles)
    self.role = (roles.flatten.map { |r| r.to_s } & LOCAL_ROLES).map { |r| 2**LOCAL_ROLES.index(r) }.sum
  end
  
  # List of roles both local and remote (no duplicates)
  def roles
    @roles ||= remote_user ? (local_roles + remote_user.roles).uniq : local_roles
  end
  
  # Check if the user has the required role
  # admin role has all privileges
  #   has_role? 'manager'
  #   has_role? :admin
  def has_role?(role)
    roles.include?("admin") || roles.include?(role.to_s)
  end

  ########################################################################
  # Method for interaction with RemoteUser
  #
  
  # Method for requesting or updating data of remote user
  REMOTE_ATTRIBUTES.each do |remote_attr| 
    # getters
    define_method remote_attr do
      remote_user.public_send(remote_attr) if remote_user
    end
    # setters
    meth = "#{remote_attr}="
    define_method meth do |val|
      # works for 1.9.2, not sure public_send() is available for 1.8.7
      remote_user.public_send(meth, val) if remote_user
    end
  end
  
  # Get the remote user via get ActiveResource
  def remote_user
    return nil unless uuid
    @remote_user ||= Rails.cache.fetch(uuid, expires_in: DisiidUser::RemoteUser.cache_expiry, race_condition_ttl: 5) do 
      DisiidUser::RemoteUser.find uuid, params: { auth_token: DisiidUser::RemoteUser.auth_token }
    end
  rescue; nil; end
    
  # Return the uuid of a user
  # Last part of identity_url, after path 'user' (element_name of RemoteUser params)
  def uuid
    @uuid ||= begin
      /.*\/#{DisiidUser::RemoteUser.collection_name}\/([a-z0-9\-]+)$/.match(identity_url.to_s)[1]
    rescue; nil; end
  end
  
  #######################################################################
  # Class RemoteUser
  # interact with remote server to get user information: personal data and roles
  # active resource
  #
  # Useful settings:
  #  - DisiidUser::RemoteUser.format = format
  #  - ActiveSupport.parse_json_times = true
  #
  # ROLES defined here must be EQUAL with the one defined remote Class self.site
  class RemoteUser < ActiveResource::Base
    self.site = 'http://localhost:3001'
    self.element_name = 'user'
    
    # authentication token that will be appended at the end of the request URL,
    # e.g. http://id.provider.com/:collection_name/:uuid?auth_token=:auth_token
    cattr_accessor :auth_token
    
    # remote_user cache expiration time, in seconds
    # defaults to 1 hour
    cattr_accessor :cache_expiry
    self.cache_expiry = 3600

    
    protected
    
    # Redefined method from Base class: we need to add auth_token suffix 
    # to the request params
    def element_path(options = nil)
      auth = { :auth_token => self.class.auth_token }
      self.class.element_path(to_param, (options || prefix_options || {}).merge(auth))
    end
  end
 
end

